unit Battery;

interface
uses ctypes, cconsts, classes, cexcepts, functions, math, scale;

const ADC_MAX = 1023;
      BYTE_MAX = 255;
      ADC_MAX_VOLT = 1800;
type
  TBattery = class;
  TBatteryClass = class of TBattery;

  TBattery =  class (TComponent)
  private
     FAlias: string;
     FFlashLenght: integer;
     fBatteryType : string; //non-edited battery type eg. NICd
     fcaps : double; //capacitance
     fstarttemp, fmaxtemp : double; //start and maximum temperature
     fcells: integer;
     fBatteryClass: TBatteryClass;
     FlashBuffer : array of integer;
     fbatsettings : TBatSettingsValues;
     ftriclechargeratio: double;
     FActivationChargeRatio: double;
     FRapidChargeRatio: double;
     FTotalTimeout: double;
     FRapidTimeout: double;
     FActivationTimeout: double;
     fDischarge: boolean;
     FDischargeStop: double;
     fRapidStartVoltage: double;
     fTricleTimeout: double;
     procedure SetAlias(const Value: string);
     function GetFlashItem(Index: integer): integer;
     procedure SetCaps(const Value: double); virtual;
     procedure SetMaxTemp(const Value: double); virtual;
     procedure SetStartTemp(const Value: double); virtual;
     procedure SetCells(const Value: integer); virtual;
     procedure SetTricleChargeRatio(const Value: double);
     procedure SetActivationChargeRatio(const Value: double);
     procedure SetTotalTimeout(const Value: double);
     procedure SetActivationTimeout(const Value: double);
     procedure SetRapidTimeout(const Value: double);
     procedure SetRapidChargeRatio(const Value: double);
     procedure SetDischargeStop(const Value: double);
    procedure SetRapidStartVoltage(const Value: double);
    procedure SetTricleTimeout(const Value: double);
   protected
     procedure ConvertSettingsToFlash;
   public
     constructor Create(AOwner : Tcomponent); override;
     destructor Destroy; override;
     function CalcBatterySettings : boolean; virtual; abstract;
     property BatteryType : string read fBatteryType;
     procedure Assign(Source: TPersistent); override;
     property BatteryClass : TBatteryClass read fBatteryClass;
     property BatterySettingsData[Index : integer]: integer read GetFlashItem;
     property BatteryFlashLenght : integer read FFlashLenght;
   published
     property Capacitance : double read fcaps write SetCaps;
     property StartTemp : double  read fstarttemp write SetStartTemp;
     property MaxTemp : double read fmaxtemp write SetMaxTemp;
     property TricleChargeRatio : double read ftriclechargeratio
       write SetTricleChargeRatio;
     property ActivationChargeRatio : double read FActivationChargeRatio
       write SetActivationChargeRatio;
     property RapidChargeRatio : double read FRapidChargeRatio
       write SetRapidChargeRatio;
     property TotalTimeout : double read FTotalTimeout write SetTotalTimeout;
     property ActivationTimeout : double read FActivationTimeout write SetActivationTimeout;
     property RapidTimeout : double  read FRapidTimeout write SetRapidTimeout;
     property Cells : integer read fcells write SetCells;
     property Alias : string read FAlias write SetAlias;
     property DischargeFirst : boolean read fDischarge write fDischarge;
     property DischargeStop : double read FDischargeStop write SetDischargeStop;
     property RapidStartVoltage : double read fRapidStartVoltage write
       SetRapidStartVoltage;
     property TricleTimeout : double read fTricleTimeout write SetTricleTimeout;
   end;



   TBatteryList = class (TList)
   private
     function GetListItem(Index: integer): TBattery;
   public
     constructor Create(); virtual;
     destructor Destroy; override;
     property Batteries[Index : integer] : TBattery read GetListItem;
   end;


   TNiCdBattery = class (TBattery)
  private
    fTempIncrease: boolean;
    fTempIncValue: double;
    fCheckVoltageDecay: boolean;
    fVoltageDecay: double;
    fDeactiveTime: double;
    procedure SetDeactiveTime(const Value: double);
    procedure SetVoltageDecay(const Value: double);
    procedure SerTempIncValue(const Value: double);
   public
     function CalcBatterySettings : boolean; override;
     constructor Create(AOwner : Tcomponent); override;
     procedure Assign(Source: TPersistent); override;
   published
     property CheckTempIncrease : boolean read fTempIncrease write fTempIncrease;
     property TempIncValue : double read fTempIncValue write SerTempIncValue;
     property CheckVoltageDecay : boolean read fCheckVoltageDecay write
       fCheckVoltageDecay;
     property VoltageDecay : double read fVoltageDecay write
       SetVoltageDecay;
     property DeactiveTime : double read fDeactiveTime
       write SetDeactiveTime;
   end;

   TNiMHBattery = class (TNiCdBattery)
     constructor Create(AOwner : Tcomponent); override;
   end;

   TLionBattery = class (TBattery)
  private
    fRapidVoltageLimit: double;
    fTricleVoltageLimit: double;
    fRapidTerminationCurrent: double;
    procedure SetRapidVoltageLimit(const Value: double);
    procedure SetRapidTerminationCurrent(const Value: double);
    procedure SetTricleVoltageLimit(const Value: double);
  public
    function CalcBatterySettings : boolean; override;
    constructor Create(AOwner : Tcomponent); override;
    procedure Assign(Source: TPersistent); override;
  published
    property RapidVoltageLimit : double read fRapidVoltageLimit
      write SetRapidVoltageLimit;
    property RapidTerminationCurrent : double read fRapidTerminationCurrent
      write SetRapidTerminationCurrent;
    property TricleVoltageLimit : double read fTricleVoltageLimit
      write SetTricleVoltageLimit;
   end;

   TSLABattery = class (TLionBattery)
   public
     constructor Create(AOwner : Tcomponent); override;
     function CalcBatterySettings : boolean; override;
   end;

   TLiPBattery = class (TLionBattery)
   public
     constructor Create(AOwner : Tcomponent); override;
   end;

implementation
{ TBattery }

procedure TBattery.Assign(Source: TPersistent);
var tmp : TBattery;
begin
  tmp := (source as TBattery);

  fBatteryType := tmp.fBatteryType;

  fcaps := tmp.fcaps;
  fstarttemp := tmp.fstarttemp;
  fmaxtemp := tmp.fmaxtemp;

  fcells := tmp.fcells;

  fBatteryClass := tmp.fBatteryClass;
  fbatsettings := tmp.fbatsettings;

  fTricleChargeRatio := tmp.fTricleChargeRatio;
  FActivationChargeRatio := tmp.FActivationChargeRatio;
  FRapidChargeRatio := tmp.RapidChargeRatio;

  FTotalTimeout := tmp.FTotalTimeout;
  FRapidTimeout := tmp.FRapidTimeout;
  FActivationTimeout := tmp.FActivationTimeout;
  fTricleTimeout := tmp.fTricleTimeout;

  Name := tmp.name;

  DischargeFirst := tmp.DischargeFirst;
  DischargeStop := tmp.DischargeStop;

  RapidStartVoltage := tmp.RapidStartVoltage;

end;

procedure TBattery.SetDischargeStop(const Value: double);
begin
  FDischargeStop := Value;
end;

procedure TBattery.ConvertSettingsToFlash;
var i : integer;
    pa : PByteArraySettings;
begin
  FFlashLenght := CFLASH_SIZE_BLOCK*
    ((sizeof(fbatsettings) div CFLASH_SIZE_BLOCK) + 1);

  SetLength(FlashBuffer, FFlashLenght);
  for i := 0 to FFlashLenght-1 do FlashBuffer[i] := 0;

  pa := PByteArraySettings(@fbatsettings);
  for i := 0 to sizeof(fbatsettings)-1 do
    FlashBuffer[i] := pa^[i];
end;

constructor TBattery.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);

  fcaps := 0;
  fstarttemp := 0;
  fmaxtemp := 0;
  fcells := 0;

  ftriclechargeratio := 0;
  FActivationChargeRatio := 0;
  FRapidChargeRatio := 0;

  FTotalTimeout := 0;
  FRapidTimeout := 0;
  FActivationTimeout := 0;
  fTricleTimeout := 0;

  fBatteryType := '';
  FAlias := 'Part #';
  fBatteryClass := TBatteryClass(ClassType);

  FFlashLenght := CBATTERY_FLASH_LEN;
  SetLength(FlashBuffer, FFlashLenght);

  DischargeFirst := false;
  DischargeStop := -1;
  RapidStartVoltage := -1;
end;



destructor TBattery.Destroy;
begin
  FlashBuffer := nil;
  inherited;
end;

function TBattery.GetFlashItem(Index: integer): integer;
begin
   result := FlashBuffer[Index] and $FF;
end;

procedure TBattery.SetActivationChargeRatio(const Value: double);
begin
  FActivationChargeRatio := Value;
end;

procedure TBattery.SetActivationTimeout(const Value: double);
begin
  FActivationTimeout := Value;
end;

procedure TBattery.SetAlias(const Value: string);
begin
  FAlias := Value;
end;

procedure TBattery.SetCaps(const Value: double);
begin
  fcaps := Range(Value, CBATCAPMIN, CBATCAPMAX);
end;

procedure TBattery.SetCells(const Value: integer);
begin
  fcells := Range(Value, NCELLSMIN, NCELLSMAX);
end;



procedure TBattery.SetMaxTemp(const Value: double);
begin
  fmaxtemp := Range(value, TMAXMIN, TMAXMAX);
end;

procedure TBattery.SetRapidChargeRatio(const Value: double);
begin
  FRapidChargeRatio := Range(Value, 0, 10);
end;

procedure TBattery.SetRapidTimeout(const Value: double);
begin
  FRapidTimeout := Range(Value, 0, 30000);
end;

procedure TBattery.SetStartTemp(const Value: double);
begin
  fstarttemp := Range(value, TMINMIN, TMINMAX);
end;

procedure TBattery.SetTotalTimeout(const Value: double);
begin
  FTotalTimeout := Value;
end;

procedure TBattery.SetTricleChargeRatio(const Value: double);
begin
  ftriclechargeratio := Value;
end;

procedure TBattery.SetRapidStartVoltage(const Value: double);
begin
  fRapidStartVoltage := Range(Value, 0, 5);
end;

procedure TBattery.SetTricleTimeout(const Value: double);
begin
  fTricleTimeout := Range(Value, 0, 30000);
end;

{ TBatteryList }

constructor TBatteryList.Create;
var bt : TBattery;
begin
  inherited;
  bt := TNiCdBattery.Create(nil);
  Add(bt);
  bt := TNiMHBattery.Create(nil);
  Add(bt);
  bt := TLionBattery.Create(nil);
  Add(bt);
  bt := TLiPBattery.Create(nil);
  Add(bt);
  bt := TSLABattery.Create(nil);
  Add(bt);
end;

destructor TBatteryList.Destroy;
var i : integer;
    tmp : TBattery;
begin
  for i := 0 to Count-1 do
  begin
    tmp := Items[i];
    tmp.Free;
  end;
  inherited;
end;

function TBatteryList.GetListItem(Index: integer): TBattery;
begin
  result :=  TBattery(Items[Index]);
end;

{ TNiCdBattery }


procedure TNiCdBattery.Assign(Source: TPersistent);
var tmp : TNiCdBattery;
begin
  inherited;

  tmp := (Source as TNiCdBattery);
  CheckTempIncrease := tmp.CheckTempIncrease;
  TempIncValue := tmp.TempIncValue;
  CheckVoltageDecay := tmp.CheckVoltageDecay;
  VoltageDecay := tmp.VoltageDecay;
  DeactiveTime := tmp.DeactiveTime;
end;

function TNiCdBattery.CalcBatterySettings: boolean;
begin
  with fbatsettings do
  begin
     discharge_before_charge := byte(DischargeFirst);
     ro(@discharge_before_charge, sizeof(discharge_before_charge));

     discharge_thl_voltage := CheckRangeEx('discharge_thl_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(DischargeStop*Cells));
     ro(@discharge_thl_voltage, sizeof(discharge_thl_voltage));

     temperature_min := CheckRangeEx('temperature_min', 0, BYTE_MAX,
       Scales[TTemperature].reversescalei(StartTemp));
     ro(@temperature_min, sizeof(temperature_min));

     temperature_max := CheckRangeEx('temperature_max', 0, BYTE_MAX,
     Scales[TTemperature].reversescalei(MaxTemp));
     ro(@temperature_max, sizeof(temperature_max));

     waiting_charge_voltage := CheckRangeEx('waiting_charge_voltage', 0, ADC_MAX_VOLT,
      Scales[TVoltage].reversescalei(NICD_TRICLE_CHARGE_VOLTAGE*Cells));
     ro(@waiting_charge_voltage, sizeof(waiting_charge_voltage));

     waiting_charge_current := CheckRangeEx('waiting_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*TricleChargeRatio));
     ro(@waiting_charge_current, sizeof(waiting_charge_current));

     tricle_charge_voltage :=  CheckRangeEx('tricle_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(NICD_TRICLE_CHARGE_VOLTAGE*Cells));
     ro(@tricle_charge_voltage, sizeof(tricle_charge_voltage));

     tricle_charge_current :=  CheckRangeEx('tricle_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*TricleChargeRatio));
     ro(@tricle_charge_current, sizeof(tricle_charge_current));

     tricle_time := Scales[TTime].reversescalei(
       TricleTimeout*SECS_PER_MINUTE/TIMER2_INTERVAL);
     ro(@tricle_time, sizeof(tricle_time));

     charge_term_voltage := CheckRangeEx('charge_term_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(NICD_BATTERY_MAX_VOLTAGE*Cells));
     ro(@charge_term_voltage, sizeof(charge_term_voltage));

     total_charge_time :=  CheckRangeEx('total_charge_time', 0, BYTE_MAX,
       Scales[TTime].reversescalei(SECS_PER_MINUTE*TotalTimeout/TIMER3_INTERVAL));
     ro(@total_charge_time, sizeof(total_charge_time));

     activation_charge_voltage :=  CheckRangeEx('activation_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(NICD_TRICLE_CHARGE_VOLTAGE*Cells));
     ro(@activation_charge_voltage, sizeof(activation_charge_voltage));

     activation_charge_current := CheckRangeEx('activation_charge_voltage', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*ActivationChargeRatio));
     ro(@activation_charge_current, sizeof(activation_charge_current));

     activatiom_time := Scales[TTime].reversescalei(
       ActivationTimeout*SECS_PER_MINUTE/TIMER2_INTERVAL);
     ro(@activatiom_time, sizeof(activatiom_time));

     rapid_charge_voltage := CheckRangeEx('rapid_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(NICD_TRICLE_CHARGE_VOLTAGE*Cells));
     ro(@rapid_charge_voltage, sizeof(rapid_charge_voltage));

     rapid_charge_current :=  CheckRangeEx('rapid_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*RapidChargeRatio));
     ro(@rapid_charge_current, sizeof(rapid_charge_current));

     rapid_start_voltage  := CheckRangeEx('rapid_start_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(RapidStartVoltage*Cells));
     ro(@rapid_start_voltage, sizeof(rapid_start_voltage));

     rapid_stop_voltage := CheckRangeEx('rapid_stop_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(NICD_RAPID_STOP_VOLTAGE*Cells));
     ro(@rapid_stop_voltage, sizeof(rapid_stop_voltage));

     rapid_time_deactive := Scales[TTime].reversescalei(
       SECS_PER_MINUTE*DeactiveTime/TIMER2_INTERVAL);
     ro(@rapid_time_deactive, sizeof(rapid_time_deactive));

     rapid_time :=
       Scales[TTime].reversescalei(SECS_PER_MINUTE*RapidTimeout/TIMER2_INTERVAL);
     ro(@rapid_time, sizeof(rapid_time));

     rapid_check_temperature_rising := byte(CheckTempIncrease);
     ro(@rapid_check_temperature_rising, sizeof(rapid_check_temperature_rising));

     rapid_temperature_rising :=
         Scales[TTemperature].reversescalei(TempIncValue*TEMP_MEASURING_POINTS_PER_MINUTE)
         - Scales[TTemperature].reversescalei(0);
     ro(@rapid_temperature_rising, sizeof(rapid_temperature_rising));

     rapid_check_voltage_decay := byte(CheckVoltageDecay);
     ro(@rapid_check_voltage_decay, sizeof(rapid_check_voltage_decay));

     rapid_voltage_decay :=
       Scales[TVoltage].reversescalei(Cells*VoltageDecay);
     ro(@rapid_voltage_decay, sizeof(rapid_voltage_decay));

     rapid_check_current_decay := byte(false);
     ro(@rapid_check_current_decay, sizeof(rapid_check_current_decay));

     rapid_current_decay := 0;
     ro(@rapid_current_decay, sizeof(rapid_current_decay));
  end;
  ConvertSettingsToFlash();
  result := true;
end;

constructor TNiCdBattery.Create;
begin
  inherited;
  fcells := 2;
  fBatteryType := 'NiCd';

  Capacitance := 0.75;

  StartTemp := 10;
  MaxTemp := 60;

  TricleChargeRatio := 0.025;
  ActivationChargeRatio := 0.3;
  RapidChargeRatio := 0.7;

  TotalTimeout := 720;
  ActivationTimeout := 120;
  RapidTimeout := 120;
  TricleTimeout := 360;

  DischargeFirst := false;
  DischargeStop := 0.8;

  CheckTempIncrease := false;
  TempIncValue := 1.0;

  CheckVoltageDecay := true;
  VoltageDecay := 0.015;

  DeactiveTime := 3;
  RapidStartVoltage:= 1.05;
end;




procedure TNiCdBattery.SerTempIncValue(const Value: double);
begin
  fTempIncValue := Range(Value, 0, 100);
end;

procedure TNiCdBattery.SetDeactiveTime(const Value: double);
begin
  fDeactiveTime := Range(Value, 0, 1000);
end;

procedure TNiCdBattery.SetVoltageDecay(const Value: double);
begin
  fVoltageDecay := Range(Value, 0, 10);
end;

{ TNiMHBattery }

constructor TNiMHBattery.Create;
begin
 inherited;
 fBatteryType := 'NiMH';
 cells := 2;

 CheckTempIncrease := true;
 CheckVoltageDecay := false;
end;

{ TLionBattery }

procedure TLionBattery.Assign(Source: TPersistent);
var tmp : TLionBattery;
begin
  inherited;
  tmp := (source as TLionBattery);
  RapidTerminationCurrent := tmp.RapidTerminationCurrent;
  RapidVoltageLimit := tmp.RapidVoltageLimit;
  TricleVoltageLimit := tmp.TricleVoltageLimit;
end;


function TLionBattery.CalcBatterySettings: boolean;
begin
 with fbatsettings do
  begin
     discharge_before_charge := byte(DischargeFirst);
     ro(@discharge_before_charge, sizeof(discharge_before_charge));

     discharge_thl_voltage := CheckRangeEx('discharge_thl_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(DischargeStop*Cells));
     ro(@discharge_thl_voltage, sizeof(discharge_thl_voltage));

     temperature_min := Scales[TTemperature].reversescalei(StartTemp);
     ro(@temperature_min, sizeof(temperature_min));

     temperature_max := Scales[TTemperature].reversescalei(MaxTemp);
     ro(@temperature_max, sizeof(temperature_max));

     waiting_charge_voltage := 0;
     ro(@waiting_charge_voltage, sizeof(waiting_charge_voltage));

     waiting_charge_current := 0;
     ro(@waiting_charge_current, sizeof(waiting_charge_current));

     tricle_charge_voltage := CheckRangeEx('tricle_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(TricleVoltageLimit*Cells));
     ro(@tricle_charge_voltage, sizeof(tricle_charge_voltage));

     tricle_charge_current := CheckRangeEx('tricle_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*TricleChargeRatio));
     ro(@tricle_charge_current, sizeof(tricle_charge_current));

     tricle_time :=
       Scales[TTime].reversescalei(TricleTimeout*SECS_PER_MINUTE/TIMER2_INTERVAL);
     ro(@tricle_time, sizeof(tricle_time));

     charge_term_voltage := CheckRangeEx('charge_term_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(LIION_CHARGE_TERM_VOLTAGE*Cells));
     ro(@charge_term_voltage, sizeof(charge_term_voltage));

     total_charge_time := CheckRangeEx('total_charge_time', 0, BYTE_MAX,
       Scales[TTime].reversescalei(SECS_PER_MINUTE*TotalTimeout/TIMER3_INTERVAL));
     ro(@total_charge_time, sizeof(total_charge_time));

     activation_charge_voltage := CheckRangeEx('activation_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(RapidVoltageLimit*Cells));
     ro(@activation_charge_voltage, sizeof(activation_charge_voltage));

     activation_charge_current :=  CheckRangeEx('activation_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*ActivationChargeRatio));
     ro(@activation_charge_current, sizeof(activation_charge_current));

     activatiom_time := Scales[TTime].reversescalei(
       ActivationTimeout*SECS_PER_MINUTE/TIMER2_INTERVAL);
     ro(@activatiom_time, sizeof(activatiom_time));

     rapid_charge_voltage :=  CheckRangeEx('rapid_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(RapidVoltageLimit*Cells));
     ro(@rapid_charge_voltage, sizeof(rapid_charge_voltage));

     rapid_charge_current := CheckRangeEx('rapid_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*RapidChargeRatio));
     ro(@rapid_charge_current, sizeof(rapid_charge_current));

     rapid_start_voltage  := CheckRangeEx('rapid_start_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(RapidStartVoltage*Cells));
     ro(@rapid_start_voltage, sizeof(rapid_start_voltage));

     rapid_stop_voltage := CheckRangeEx('rapid_stop_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(LIION_CHARGE_TERM_VOLTAGE*Cells));
     ro(@rapid_stop_voltage, sizeof(rapid_stop_voltage));

     rapid_time_deactive :=
       Scales[TTime].reversescalei(
       SECS_PER_MINUTE*LIION_CKECKINGS_DEACTIVE_TIME/TIMER2_INTERVAL);
     ro(@rapid_time_deactive, sizeof(rapid_time_deactive));

     rapid_time :=
       Scales[TTime].reversescalei(SECS_PER_MINUTE*RapidTimeout/TIMER2_INTERVAL);
     ro(@rapid_time, sizeof(rapid_time));

     rapid_check_temperature_rising := byte(false);
     ro(@rapid_check_temperature_rising, sizeof(rapid_check_temperature_rising));

     rapid_temperature_rising := 0;
     ro(@rapid_temperature_rising, sizeof(rapid_temperature_rising));

     rapid_check_voltage_decay := byte(false);
     ro(@rapid_check_voltage_decay, sizeof(rapid_check_voltage_decay));

     rapid_voltage_decay := 0;
     ro(@rapid_voltage_decay, sizeof(rapid_voltage_decay));

     rapid_check_current_decay := byte(true);
     ro(@rapid_check_current_decay, sizeof(rapid_check_current_decay));

     rapid_current_decay := CheckRangeEx('rapid_current_decay', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*RapidTerminationCurrent));
     ro(@rapid_current_decay, sizeof(rapid_current_decay));
  end;
  ConvertSettingsToFlash();
  result := true;
end;

constructor TLionBattery.Create(AOwner : Tcomponent);
begin
  inherited;
  fBatteryType := 'LiIon';

  ActivationChargeRatio := 0.1;
  ActivationTimeout := 120.0;

  Capacitance := 0.9;
  Cells := 1;

  DischargeFirst := false;
  DischargeStop := 3.4;

  TotalTimeout := 720;

  StartTemp := 10;
  MaxTemp := 45;

  RapidStartVoltage := 3.0;
  RapidVoltageLimit := 4.1;
  RapidChargeRatio  := 0.7;
  RapidTerminationCurrent := 0.1;
  RapidTimeout := 120.0;

  TricleVoltageLimit := 4.1;
  TricleChargeRatio := 0.01;
  TricleTimeout := 5;
end;


procedure TLionBattery.SetRapidTerminationCurrent(const Value: double);
begin
  fRapidTerminationCurrent := Range(Value, 0, 10);
end;

procedure TLionBattery.SetRapidVoltageLimit(const Value: double);
begin
  fRapidVoltageLimit := Range(Value, 0, 10);
end;

procedure TLionBattery.SetTricleVoltageLimit(const Value: double);
begin
  fTricleVoltageLimit := Range(Value, 0, 10);
end;

{ TSLABattery }

function TSLABattery.CalcBatterySettings: boolean;
begin
 with fbatsettings do
  begin
     discharge_before_charge := byte(DischargeFirst);
     ro(@discharge_before_charge, sizeof(discharge_before_charge));

     discharge_thl_voltage :=  CheckRangeEx('discharge_thl_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(DischargeStop*Cells));
     ro(@discharge_thl_voltage, sizeof(discharge_thl_voltage));

     temperature_min := Scales[TTemperature].reversescalei(StartTemp);
     ro(@temperature_min, sizeof(temperature_min));

     temperature_max := Scales[TTemperature].reversescalei(MaxTemp);
     ro(@temperature_max, sizeof(temperature_max));

     waiting_charge_voltage := 0;
     ro(@waiting_charge_voltage, sizeof(waiting_charge_voltage));

     waiting_charge_current := 0;
     ro(@waiting_charge_current, sizeof(waiting_charge_current));

     tricle_charge_voltage := CheckRangeEx('tricle_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(TricleVoltageLimit*Cells));
     ro(@tricle_charge_voltage, sizeof(tricle_charge_voltage));

     tricle_charge_current := CheckRangeEx('tricle_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*TricleChargeRatio));
     ro(@tricle_charge_current, sizeof(tricle_charge_current));

     tricle_time :=
       Scales[TTime].reversescalei(TricleTimeout*SECS_PER_MINUTE/TIMER2_INTERVAL);
     ro(@tricle_time, sizeof(tricle_time));

     charge_term_voltage :=
       Scales[TVoltage].reversescalei(SLA_CHARGE_TERM_VOLTAGE*Cells);
     ro(@charge_term_voltage, sizeof(charge_term_voltage));

     total_charge_time := CheckRangeEx('total_charge_time', 0, BYTE_MAX,
       Scales[TTime].reversescalei(SECS_PER_MINUTE*TotalTimeout/TIMER3_INTERVAL));
     ro(@total_charge_time, sizeof(total_charge_time));

     activation_charge_voltage := CheckRangeEx('activation_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(RapidVoltageLimit*Cells));
     ro(@activation_charge_voltage, sizeof(activation_charge_voltage));

     activation_charge_current := CheckRangeEx('activation_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*ActivationChargeRatio));
     ro(@activation_charge_current, sizeof(activation_charge_current));

     activatiom_time := Scales[TTime].reversescalei(
       ActivationTimeout*SECS_PER_MINUTE/TIMER2_INTERVAL);
     ro(@activatiom_time, sizeof(activatiom_time));

     rapid_charge_voltage := CheckRangeEx('rapid_charge_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(RapidVoltageLimit*Cells));
     ro(@rapid_charge_voltage, sizeof(rapid_charge_voltage));

     rapid_charge_current := CheckRangeEx('rapid_charge_current', 0, ADC_MAX,
       Scales[TCurrent].reversescalei(Capacitance*RapidChargeRatio));
     ro(@rapid_charge_current, sizeof(rapid_charge_current));

     rapid_start_voltage  := CheckRangeEx('rapid_start_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(RapidStartVoltage*Cells));
     ro(@rapid_start_voltage, sizeof(rapid_start_voltage));

     rapid_stop_voltage :=  CheckRangeEx('rapid_stop_voltage', 0, ADC_MAX_VOLT,
       Scales[TVoltage].reversescalei(SLA_CHARGE_TERM_VOLTAGE*Cells));
     ro(@rapid_stop_voltage, sizeof(rapid_stop_voltage));

     rapid_time_deactive :=
       Scales[TTime].reversescalei(
       SECS_PER_MINUTE*SLA_CKECKINGS_DEACTIVE_TIME/TIMER2_INTERVAL);
     ro(@rapid_time_deactive, sizeof(rapid_time_deactive));

     rapid_time :=
       Scales[TTime].reversescalei(SECS_PER_MINUTE*RapidTimeout/TIMER2_INTERVAL);
     ro(@rapid_time, sizeof(rapid_time));

     rapid_check_temperature_rising := byte(false);
     ro(@rapid_check_temperature_rising, sizeof(rapid_check_temperature_rising));

     rapid_temperature_rising := 0;
     ro(@rapid_temperature_rising, sizeof(rapid_temperature_rising));

     rapid_check_voltage_decay := byte(false);
     ro(@rapid_check_voltage_decay, sizeof(rapid_check_voltage_decay));

     rapid_voltage_decay := 0;
     ro(@rapid_voltage_decay, sizeof(rapid_voltage_decay));

     rapid_check_current_decay := byte(true);
     ro(@rapid_check_current_decay, sizeof(rapid_check_current_decay));

     rapid_current_decay :=
       Scales[TCurrent].reversescalei(Capacitance*RapidTerminationCurrent);
     ro(@rapid_current_decay, sizeof(rapid_current_decay));
  end;
  ConvertSettingsToFlash();
  result := true;
end;

constructor TSLABattery.Create(AOwner : Tcomponent);
begin
  inherited;
  fBatteryType := 'SLA';

  ActivationChargeRatio := 0.1;
  ActivationTimeout := 120.0;

  Capacitance := 1.2;
  Cells := 3;

  DischargeFirst := false;
  DischargeStop := 2;

  TotalTimeout := 720;

  StartTemp := 0;
  MaxTemp := 45;

  RapidStartVoltage := 2.2;
  RapidVoltageLimit := 2.45;
  RapidChargeRatio  := 0.7;
  RapidTerminationCurrent := 0.2;
  RapidTimeout := 120.0;

  TricleVoltageLimit := 2.25;
  TricleChargeRatio := 0.1;
  TricleTimeout := 30;
end;

{ TLiPBattery }


constructor TLiPBattery.Create(AOwner : Tcomponent);
begin
  inherited;
  fBatteryType := 'LiPol';
  RapidChargeRatio := 0.5;
end;

end.
