unit Termistor;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, ValEdit, StdCtrls, NumInp12, ExtCtrls, xyGraph, Menus,
  cconsts, ctypes, math, lm;

type TPointArray = array of record t, r : double; end;

type
  TfrmTermistor = class(TForm)
    rTerms: TNumInp;
    Label1: TLabel;
    ValueListEditor: TValueListEditor;
    btnGenerate: TButton;
    btnClose: TButton;
    Label2: TLabel;
    xyGraph: TxyGraph;
    PopupMenuList: TPopupMenu;
    Load1: TMenuItem;
    Save1: TMenuItem;
    OpenTDialog: TOpenDialog;
    SaveTDialog: TSaveDialog;
    PopupMenuGraph: TPopupMenu;
    Yscale: TMenuItem;
    procedure btnCloseClick(Sender: TObject);
    procedure btnGenerateClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PopupMenuGraphPopup(Sender: TObject);
    procedure YscaleClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure Load1Click(Sender: TObject);
    procedure Save1Click(Sender: TObject);
  private
    { Private declarations }
    ValuesArray: TPointArray;
    PointCount : integer;
    fAddMessage: TAddMessage;
    ApproxParams, ApproxParamsCoarse : VectorDouble;
    fCalibrationPresent : boolean;
    fOnMenuEnable: TNotifyEvent;
    function FillPointArray : boolean;
    procedure AddPointsToGraph();
    function AproximateValues() : boolean;
    function getr(cp: array of double; const t: double): double;
    function gettemp(const c: array of double; const r: double): double;
    function CubicSolve(var X: array of double; a, b, c: double): integer;
    function GetCalibrationCount: integer;
    function GetCalibrationData(Index: integer): integer;
    procedure PrintStatistic();
  public
    { Public declarations }
    property PrintMessage : TAddMessage read fAddMessage write fAddMessage;
    property CalibrationData[Index : integer] : integer read GetCalibrationData;
    property CalibrationDataCount : integer read GetCalibrationCount;
    property CalibrationPresent : boolean read fCalibrationPresent;
    property OnMenuEnable : TNotifyEvent read fOnMenuEnable write fOnMenuEnable;
  end;

var
  frmTermistor: TfrmTermistor;

implementation

{$R *.dfm}

//uses Cardano formulas
function TfrmTermistor.CubicSolve(var X : array of double; a, b, c : double) : integer;
var q,r,r2,q3,t,aa,bb : double;
begin
  q:=(a*a-3.0*b)/9.0;
  r:=(a*(2.0*a*a-9.0*b)+27.0*c)/54.0;
  r2:=r*r; q3:=q*q*q;
  if(r2<q3) then
  begin
    t := arccos(r/sqrt(q3));
    a := a/3.0;
    q := -2.0*sqrt(q);
    x[0] :=q*cos(t/3.0)-a;
    x[1] :=q*cos((t+2*pi)/3.0)-a;
    x[2] :=q*cos((t-2*pi)/3.0)-a;
    result := 3;
  end
  else
  begin

    if (r = 0) then aa := 0 else
    if (r > 0) then aa := -power(r+sqrt(r2-q3),1/3)
               else aa :=  power(-r+sqrt(r2-q3),1/3);
    

    if(aa <> 0) then bb := q/aa else bb := 0;

    a := a/3; q:=aa+bb; r:=aa-bb;
    x[0] :=q-a;
    x[1] :=(-0.5)*q-a;
    x[2] :=(sqrt(3)*0.5)*abs(r);
    if(x[2]=0) then result := 2 else result := 1;
  end;

end;

// calculates the termistor temperature for given termistor resistance
function TfrmTermistor.gettemp(const c: array of double; const r: double): double;
var  t : double;
begin
  t := ln(r);
  result := 1/(c[1]+c[2]*t+c[3]*t*t*t)-CDELTA_TEMP;
end;

//finding the termistor resistance for given temparature
function TfrmTermistor.getr(cp : array of double; const t : double) : double;
const CMAX_DISTANCE : double = 1e100;
var a, b, c, re, distance, delta, rx  : double;
    x : array [0..2] of double;
    nr, i, imin : integer;
begin

  a := 0;
  b := cp[2]/cp[3];
  c := (cp[1] - 1/(t + CDELTA_TEMP))/cp[3];

  nr := CubicSolve(x, a, b, c);

  if (nr = 1) then re := x[0] else
  begin
    distance := CMAX_DISTANCE;
    imin := 0;
    rx := (1/(t + CDELTA_TEMP) - ApproxParamsCoarse[1])/ApproxParamsCoarse[2];
    for i := 0 to nr-1 do
    begin
      delta := abs(rx-x[i]);
      if (delta< distance) then
      begin
        distance := delta;
        imin := i;
      end;
    end;
    re := x[imin];
  end;

 result := exp(re);
end;

procedure TfrmTermistor.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmTermistor.btnGenerateClick(Sender: TObject);
begin
  VCHARGER_RVALS.Rterms := rTerms.Value;

  if  FillPointArray and
    AproximateValues then
  begin
    AddPointsToGraph();
    PrintStatistic();
    fCalibrationPresent := true;
    OnMenuEnable(self);
  end
    else PrintMessage(TMError, 'It is insufficient termistor data points to calibration!')
end;

function TfrmTermistor.FillPointArray : boolean;
var i, max_rows : integer;
begin
  PointCount := 0;
  max_rows := ValueListEditor.RowCount-1;

  SetLength(ValuesArray, max_rows+1);

  for i := 1 to max_rows do
  try
    ValuesArray[PointCount].t := StrToFloat(ValueListEditor.Cells[0,i]);
    ValuesArray[PointCount].r := StrToFloat(ValueListEditor.Cells[1,i]);
    inc(PointCount);
  except
    if Assigned(Printmessage) then PrintMessage(TMError,
      Format('Invalid numerical format at %d row', [i]));
  end;

  result := PointCount > CN_APROX_PARAMS-1;
end;

procedure TfrmTermistor.FormDestroy(Sender: TObject);
begin
  ValuesArray := nil;
  ApproxParams := nil;
  ApproxParamsCoarse := nil;
end;

procedure TfrmTermistor.FormCreate(Sender: TObject);
var I : integer;
begin
  PrintMessage := nil;
  SetLength(ApproxParams, CN_APROX_PARAMS+1);
  SetLength(ApproxParamsCoarse, CN_APROX_PARAMS+1);;

  for i := 1 to CN_APROX_PARAMS do
  begin
    ApproxParams[i] := 0;
    ApproxParamsCoarse[i] := 0;
  end;  

  fCalibrationPresent := false;

  OpenTDialog.DefaultExt := 'txt';
  OpenTDialog.Filter := 'Text files (*.txt)|*.txt';
  OpenTDialog.InitialDir := GetCurrentDir;

  SaveTDialog.DefaultExt := 'txt';
  SaveTDialog.Filter := 'Text files (*.txt)|*.txt';
  SaveTDialog.InitialDir := GetCurrentDir;

end;

procedure TfrmTermistor.AddPointsToGraph;
var i : integer;
    step, t, r : double;
begin
  with xygraph do
  begin
    ClearAll;
    Plotting := false;

    XAxis.AutoSizing := true;
    YAxis.AutoSizing := true;
    XAxis.AutoStepping := true;
    YAxis.AutoStepping := true;
    YAxis.LogScale := true;

    Series[1].DrawLine := true;
    Series[1].DrawPoints := false;
    Series[1].LineStyle := psSolid;
    Series[1].LineColor := clGreen;

    Series[2].DrawLine := false;
    Series[2].DrawPoints := true;
    Series[2].PointShape := ps_Square;
    Series[2].PointColor := clBlue;

    step := (CMAX_TEMPERATURE_GRAPH - CMIN_TEMPERATURE_GRAPH)/CN_TAB_GRAPH_POINTS;

    for i := 0 to CN_TABCHARGER_POINTS-1 do
    begin
      t := CMIN_TEMPERATURE_GRAPH + i*step;
      r := getr(ApproxParams, t);
      Series[1].Add(t, r);
    end;

    for i := 0 to PointCount-1 do
      Series[2].Add(ValuesArray[i].t, ValuesArray[i].r);

    Plotting :=True;
  end;
end;

function TfrmTermistor.AproximateValues : boolean;
var
  chisq : double;
  i: integer;
  x,y,sig : VectorDouble;
  init : VectorDouble;
  fix_vals : VectorInt;
begin
  result := false;

  try
    SetLength(x, PointCount+1);
    SetLength(y, PointCount+1);
    SetLength(sig, PointCount+1);

    SetLength(init, CN_APROX_PARAMS+1);
    SetLength(fix_vals, CN_APROX_PARAMS+1);

    init[1] := 1e-5;
    init[2] := 1e-5;
    init[3] := 0;

  for i := 1 to CN_APROX_PARAMS do fix_vals[i] := 1;

  for i := 0 to PointCount-1 do
  begin
    x[i+1] := ValuesArray[i].r;
    y[i+1] := ValuesArray[i].t;
    sig[i+1] := 1;
  end;

    try

      fix_vals[3] := 0;

      result := MinimizeFunct(x, y, sig, PointCount, funcs, CN_APROX_PARAMS,
      CN_APROX_PARAMS-1, fix_vals, init, ApproxParamsCoarse, chisq);
      // calculates the parameter initial values

      if (result) then
      begin
        for i := 1 to CN_APROX_PARAMS do
          ApproxParams[i] := ApproxParamsCoarse[i];
        fix_vals[3] := 1;

        //refine it here via non-linear optimization
        result := MinimizeFunct(x, y, sig, PointCount, funcs, CN_APROX_PARAMS,
           CN_APROX_PARAMS, fix_vals, init, ApproxParams, chisq);
      end;


    finally
      x := nil; y := nil; sig := nil; init := nil; fix_vals := nil;
    end;

  except
    on E: Exception do PrintMessage(TMError,E.Message);
  end;

  if result then PrintMessage(TMInfo, 'The parameters were well calculated!');
end;


procedure TfrmTermistor.PopupMenuGraphPopup(Sender: TObject);
begin
  if xyGraph.YAxis.LogScale then Yscale.Caption := 'Y Scale Linear'
                            else Yscale.Caption := 'Y Scale Log';
end;

procedure TfrmTermistor.YscaleClick(Sender: TObject);
begin
  xyGraph.YAxis.LogScale := not xyGraph.YAxis.LogScale;  
end;

function TfrmTermistor.GetCalibrationCount: integer;
begin
  if CalibrationPresent then result := CN_TABCHARGER_POINTS else result := 0;
end;

function TfrmTermistor.GetCalibrationData(Index: integer): integer;
var t, rr, step : double;
begin
  if (Index < CN_TABCHARGER_POINTS) then
  begin
     step := (CMAX_TEMPERATURE_CHARGER - CMIN_TEMPERATURE_CHARGER)/CN_TABCHARGER_POINTS;
     t := CMIN_TEMPERATURE_CHARGER + Index*step;
     rr := getr(ApproxParams, t)/VCHARGER_RVALS.Rterms;

     result := round(2*CSD_SCALE_VALUE/(1+rr));

  end
   else raise Exception.Create('Invalid index under GetCalibrationData');

end;

procedure TfrmTermistor.PrintStatistic;
var i : integer;
    max_err, RMS_err, cur_err : double;
begin
  max_err := 0;
  RMS_err := 0;

  for i := 0 to PointCount-1 do
  begin
    cur_err := abs(getr(ApproxParams, ValuesArray[i].t)-ValuesArray[i].r)/
      ValuesArray[i].r;
    if cur_err > max_err then max_err := cur_err;
    RMS_err := cur_err*cur_err;
  end;

  max_err := 100*max_err;
  RMS_err := 100*sqrt(RMS_err)/PointCount;

  if getr(ApproxParams, CMAX_CHARGER_OPERATION_TEMPERATURE)
    < VCHARGER_RVALS.Rterms then
  PrintMessage(TMWarning, 'Please decrease divider value, ADC can be saturated!')
    else PrintMessage(TMInfo,
      Format('MAX err. %2.1g%%, RMS err. %2.1g%%', [max_err, RMS_err]));

  PrintMessage(TMInfo, Format('The estimeted parameter values are: ' +
    'A=%3.7g, B=%3.7g, C=%3.7g',
    [ApproxParams[1], ApproxParams[2], ApproxParams[3]]));

  //only basics data
end;

procedure TfrmTermistor.FormActivate(Sender: TObject);
begin
  rTerms.Value := VCHARGER_RVALS.Rterms;
end;

procedure TfrmTermistor.Load1Click(Sender: TObject);
begin
  if OpenTDialog.Execute then
    ValueListEditor.Strings.LoadFromFile(OpenTDialog.FileName);
end;

procedure TfrmTermistor.Save1Click(Sender: TObject);
begin
  if SaveTDialog.Execute then
    ValueListEditor.Strings.SaveToFile(SaveTDialog.FileName);
end;

end.
