unit commun;

interface

uses ctypes, cconsts, classes, cexcepts, functions, sysutils, cport, graphics,
     windows, battery, termistor, scale;

const N_WAIT_OBJECTS = 2;
      RXCOMPLETE = 1;
      WAKEUP     = 0;

type TCommunicator = class (TThread)
  private
    fMessage: TAddMessage;
    fmsgtype : TMessageType;
    fmsg : string;

    ComPort : TComPort;
    DataPacket : TComDataPacket;
    frxstr : string;
    com_state :  TComState;
    csRxStr : TRTLCriticalSection;

    csState : TRTLCriticalSection;
    hWakeUp : THandle;
    hRxComplete : THandle;
    hSleepConfirm : THandle;
    wait_handle_arr : array[0..N_WAIT_OBJECTS-1] of THandle;

    graph_obj : TAddPoint;
    graph_point_x, graph_point_y : double;
    fAdd_U: TAddPoint;
    fAdd_I: TAddPoint;
    fAdd_T: TAddpoint;

    fListValue : double;
    fListItem : TListNames;
    fListUpdateEvent : TListItemChange;

    fChargerQueryPeriod : integer;
    fChargerWaitAnswer : integer;
    csTimeouts : TRTLCriticalSection;
    fTermistor: TfrmTermistor;

    block_index : integer;
    FFlashLenght: integer;

    FFlashBuffer : array of integer;
    fFlashStartBlock: integer;

    function ConvStr(const str: string; const left, right: integer): integer;

    procedure PrintMessage(const mestype : TMessageType; const str : string);
    procedure SynchUpdateMsg;

    procedure AddPoint(graph : TAddPoint; const x, y : double);
    procedure SynchUpdateGraph;

    procedure UpdateListValue(const Item : TListNames; const Value : double);
    procedure SynchUpdateList;

    function GetStateVar : TComState;
    procedure SetStateVar(const st : TComState);
    property State : TComState read GetStateVar write SetStateVar;
    function GetChargerQueryPeriod: integer;
    function GetChargerWaitAnswer: integer;
    procedure SetChargerQueryPeriod(const Value: integer);
    procedure SetChargerWaitAnswer(const Value: integer);
    procedure SetFlashLenght(const Value: integer);
    function GetFlashItem(Index: integer): integer;
    procedure SetFlashBuffer(Index: integer; const Value: integer);
    function GetRxStr: string;
  private
    procedure SetRxStr(const str : string);
    procedure OnPacket(Sender : TObject; const str : string);
    function GetComStr(const state : TComState) : string;
    function AnalyzeMessage(const copy_state : TComState) : boolean;
    property RxStr : string read GetRxStr write SetRxStr;
  protected
    procedure Execute; override;
  public
    constructor Create();
    destructor Destroy; override;
    procedure SetComPort();
    property MessageFunct : TAddMessage read fMessage write fMessage;
    property AddPoint_U : TAddPoint read fAdd_U write fAdd_U;
    property AddPoint_I : TAddPoint read fAdd_I write fAdd_I;
    property AddPoint_T : TAddpoint read fAdd_T write fAdd_T;
    property OnUpdateListValue : TListItemChange read fListUpdateEvent
      write fListUpdateEvent;
    function SetState(const new_state : TComState) : boolean;
    property ChargerWaitAnswer : integer read GetChargerWaitAnswer
      write SetChargerWaitAnswer;
    property ChargerQueryPeriod : integer read GetChargerQueryPeriod
      write SetChargerQueryPeriod;
    property FlashLenght : integer read FFlashLenght write SetFlashLenght;
    property FlashBuffer[Index : integer]: integer read GetFlashItem
      write SetFlashBuffer;
    property FlashStartBlock : integer read fFlashStartBlock
      write fFlashStartBlock;  
end;

implementation

{ TCommunicator }

constructor TCommunicator.Create;
begin
  inherited Create(True);

  ComPort := TComPort.Create(nil);

  try
    ComPort.LoadSettings(stRegistry, REG_COM_PORT);
  except

  end;

  DataPacket := TComDataPacket.Create(nil);
  DataPacket.OnPacket := OnPacket;
  DataPacket.ComPort := ComPort;
  DataPacket.IncludeStrings := false;
  DataPacket.CaseInsensitive := false;
  DataPacket.Size := CMAX_PACKET_SIZE;
  Datapacket.StartString := CCOM_START;
  DataPacket.StopString := CCOM_STOP;

  fMessage := nil;
  fAdd_U := nil;
  fAdd_I := nil;
  fAdd_T := nil;

  fTermistor := nil;

  fListUpdateEvent := nil;

  graph_obj := nil;

  InitializecriticalSection(csState);
  InitializeCriticalSection(csTimeouts);
  InitializeCriticalSection(csRxStr);

  hRxComplete := CreateEvent(nil, FALSE, FALSE, nil);
  hWakeUp := CreateEvent(nil, FALSE, FALSE, nil);
  hSleepConfirm :=CreateEvent(nil, FALSE, FALSE, nil);

  wait_handle_arr[RXCOMPLETE] := hRxComplete;
  wait_handle_arr[WAKEUP] := hWakeUp;

  State := SReady;
  Priority := tpNormal;
  Resume;
end;


destructor TCommunicator.Destroy;
begin
  if ComPort.Connected then ComPort.Close;

  DeleteCriticalSection(csState);
  DeleteCriticalSection(csTimeouts);
  DeleteCriticalSection(csRxStr);

  CloseHandle(hRxComplete);
  CloseHandle(hWakeUp);
  CloseHandle(hSleepConfirm);

  FFlashBuffer := nil;
end;

procedure TCommunicator.SetComport;
begin
  if ComPort.Connected then ComPort.Close;
  ComPort.ShowSetupDialog();
  ComPort.StoreSettings(stRegistry, REG_COM_PORT);
end;


function TCommunicator.SetState(const new_state: TComState): boolean;
begin
  result := true;
  ResetEvent(hSleepConfirm);

  if (State <> SReady) then
  begin
    SetEvent(hWakeUp);
    sleep(200);
    SetEvent(hWakeUp);
    if (WaitForSingleObject(hSleepConfirm, CWAIT_WAKEUP) <> WAIT_OBJECT_0)
      then
      begin
        MessageFunct(TMWarning, 'The receiver thread is busy, please repeat command');
        exit;
      end;
  end;
  //if receiver thread is busy, abort operation and switch to ready state

  if (new_state <> SReady) then
  begin

    State := new_state;
    SetEvent(hWakeUp);
  end;
  //if new_state is not Ready, send message to thread
end;


function TCommunicator.GetComStr(const state: TComState): string;
const BYTE_MASK : integer = $FF;
var tmp : string;
    i : integer;
begin
  tmp := '';
  case state of
    SRtoFlashStart:
    begin
      block_index := FlashLenght div CFLASH_SIZE_BLOCK;
      tmp := IntTohex(FlashStartBlock, 2) + IntToHex(block_index, 2);
    end;
    SRToWriteBlock:
    begin
      dec(block_index);
      for i := 0 to CFLASH_SIZE_BLOCK-1 do
      tmp := tmp + IntToHex(FlashBuffer[CFLASH_SIZE_BLOCK*block_index+i]
        and BYTE_MASK, 2);
    end;
  end;

  result := CCOM_START + SetStateMsg[state] + tmp + CCOM_STOP;
end;

procedure TCommunicator.Execute;
var cur_state, next_state : TComState;
    res : DWORD;
    v1, v2 : boolean;
begin

  while (not Terminated) do
  begin
  try
    cur_state := State;
    next_state := SReady;

    ResetEvent(hWakeUp);

    if (cur_state <> SReady) then
    begin
      ComPort.Open;
      ComPort.ClearBuffer(true, false);
      ComPort.WriteStr(GetComStr(cur_state));

      res := WaitForMultipleObjects(N_WAIT_OBJECTS, @wait_handle_arr,
          FALSE, GetChargerWaitAnswer);

      case res of
        WAIT_OBJECT_0 + RXCOMPLETE :
        begin
          v1 := AnalyzeMessage(cur_state);
          if (v1 and CWANT_TIMEOUT[cur_state]) then
              v2 := (WaitForSingleObject(hWakeup, GetChargerQueryPeriod) = WAIT_TIMEOUT)
              else v2 := true;
          if (v1 and v2) then next_state := CNEXT_STATE[cur_state];
        end;
          WAIT_TIMEOUT : PrintMessage(TMError, 'Charger is not responding.');
      end;
    end;

    State := next_state;

    if (next_state = SReady) then
    begin
      ComPort.Close;
      SetEvent(hSleepConfirm);
      WaitForSingleObject(hWakeUp, INFINITE);
    end;

  except
     on E: Exception do
     begin
       PrintMessage(TMError, E.Message);
       State := SReady;
     end;  
  end;
  end;
end;

function TCommunicator.AnalyzeMessage(const copy_state: TComState): boolean;
var state_byte : integer;
    time, i, u, t : double;
    index, tmp, code, error : integer;
    val : integer;
begin

  result := true;

  try
    state_byte := ConvStr(rxstr, TYPE_BYTE_START, TYPE_BYTE_STOP);

    case state_byte of
      CHARGER_ADD_CURRDATA :
      begin
       time := Scales[TTime].scaled(
          ConvStr(rxstr, TIME_START_BYTE, TIME_STOP_BYTE));

        val := ConvStr(rxstr, I_START_BYTE, I_STOP_BYTE);
        if ((val and $8000) > 0) then val := Longword(val) or $FFFF0000;
        //sign extension
        i := Scales[TCurrent].scaled(val);


        val := ConvStr(rxstr, U_START_BYTE, U_STOP_BYTE);
        if ((val and $8000) > 0) then val := Longword(val) or $FFFF0000;
        //sign extension
        u := Scales[TVoltage].scaled(val);

        t := Scales[TTemperature].scaled(
         ConvStr(rxstr, T_START_BYTE, T_STOP_BYTE));

        AddPoint(AddPoint_U, time, u);
        Addpoint(AddPoint_I, time, i);
        AddPoint(AddPoint_T, time, t);

        UpdateListValue(TTime, time);
        UpdateListValue(TVoltage, u);
        UpdateListValue(TCurrent, i);
        UpdateListValue(TTemperature, t);

        val := ConvStr(rxstr, CODE_BYTE_START, CODE_BYTE_STOP);
        if val = CSTOPPED_VALUE then
        begin
          PrintMessage(TMInfo, 'The charger went to STOPPED state!');
          result := false;
        end;
        UpdateListValue(TChState, val);

        error := ConvStr(rxstr, ERROR_BYTE_START, ERROR_BYTE_STOP);
        UpdateListValue(TError, error);

      end;

      CHARGER_ERROR_BATTERY : PrintMessage(TMError,'The charger reported an battery error.');

      CHARGER_ERROR_CONTROL : PrintMessage(TMError, 'The charger reported an control system error.');

      CHARGER_CONFDIAG : PrintMessage(TMInfo, 'Diagnostic responce received.');

      CHARGER_CONFSTOP : PrintMessage(TMInfo, 'The current operation was aborted, charger reset.');

      CHARGER_CONFSTART : PrintMessage(TMInfo, 'The charging start was confirmed.');

      CHARGER_DISCHARGING_COMPLETE : PrintMessage(TMInfo, 'The discharging complete.');

      CHARGER_DISCHARGING_STOP : PrintMessage(TMWarning, 'The discharging aborted.');

      CHARGER_CHARGING_COMPLETE:
      begin
        time := Scales[TTime].scaled(
          ConvStr(rxstr, TIME_START_BYTE, TIME_STOP_BYTE));

        UpdateListValue(TTime, time);

        result := false; // we must goto ready state after it
        PrintMessage(TMInfo, 'Charging process completed.');
      end;

      CHARGER_FLASH_START:
      begin
        PrintMessage(TMInfo, 'The flash writing initiated!');
        tmp := ConvStr(rxstr, CODE_BYTE_START, CODE_BYTE_STOP);
        PrintMessage(TMInfo, Format('The PSoC temperature is %dC', [tmp-$7F]));
      end;

      CHARGER_FLASH_WRITING:
      begin
        code := ConvStr(rxstr, CODE_BYTE_START, CODE_BYTE_STOP);
        index := ConvStr(rxstr, GDATA_BYTE_START, GDATA_BYTE_STOP);

        case code of
          CWRITE_FLASH_OK: PrintMessage(TMInfo,Format('The %d flash block has been written well',
          [index]));
          CWRITE_FLASH_ERROR:
          begin
            PrintMessage(TMError, Format('Error writing %d block', [index]));
            result := false;
          end;
          CWRITE_FLASH_END:
          begin
             PrintMessage(TMInfo,Format('The %d flash block has been written well',
             [index]));
             PrintMessage(TMInfo, 'Finishing writing');
             result := false;
          end
         else
          begin
           result := false;
           PrintMessage(TMError, 'Unknown report code during flash writing.');
          end;
        end;
      end
    else
      begin
        PrintMessage(TMError, 'Unknown answer has been received from charger.');
        result := false;
      end
    end;

  except
    on EComponentError do PrintMessage (TMError, 'The charger answer lenght is bad.');
    on EConvertError do PrintMessage(TMError, 'The invalid numeric value in the charger answer.');
    else PrintMessage(TMError, 'The  charger message cannot be parsed.');
    result := false;
  end;
end;


procedure TCommunicator.SetRxStr(const str: string);
begin
  EnterCriticalSection(csRxStr);
  frxstr := str;
  LeaveCriticalSection(csRxStr);
end;

function TCommunicator.ConvStr(const str: string; const left,
  right: integer): integer;
var l : integer;
    tmp : string;
begin
  l := Length(str);

  if ((l < left) or (l < right))
  then raise EComponentError.Create('The string lenght less then index');

  tmp := Copy(str, 2*left+1, 2*(right-left)+2);
  result := StrToInt('$'+tmp);
end;

procedure TCommunicator.PrintMessage(const mestype: TMessageType;
  const str : string);
begin
  fmsgtype := mestype;
  fmsg := str;
  Synchronize(SynchUpdateMsg);
end;

procedure TCommunicator.SynchUpdateMsg;
begin
  if Assigned(MessageFunct) then MessageFunct(fmsgtype, fmsg);
end;

function TCommunicator.GetStateVar: TComState;
begin
  EnterCriticalSection(csState);
  result := com_state;
  LeaveCriticalSection(csState);
end;

procedure TCommunicator.SetStateVar(const st: TComState);
begin
  EnterCriticalSection(csState);
  com_state := st;
  LeaveCriticalSection(csState);
end;

procedure TCommunicator.AddPoint(graph: TAddPoint; const x, y: double);
begin
  graph_obj := graph;
  graph_point_x :=  x;
  graph_point_y := y;
  Synchronize(SynchUpdateGraph);
end;

procedure TCommunicator.SynchUpdateGraph;
begin
  if Assigned(graph_obj) then graph_obj(graph_point_x, graph_point_y);
end;

procedure TCommunicator.SynchUpdateList;
begin
 if Assigned(fListUpdateEvent) then fListUpdateEvent(fListItem, fListValue);
end;

procedure TCommunicator.UpdateListValue(const Item: TListNames;
  const Value: double);
begin
  fListItem := Item;
  fListValue := Value;

  Synchronize(SynchUpdateList);
end;

function TCommunicator.GetChargerQueryPeriod: integer;
begin
  EnterCriticalSection(csTimeouts);
  result := fChargerQueryPeriod;
  LeaveCriticalSection(csTimeouts);
end;

function TCommunicator.GetChargerWaitAnswer: integer;
begin
  EnterCriticalSection(csTimeouts);
  result := fChargerWaitAnswer;
  LeaveCriticalSection(csTimeouts);
end;

procedure TCommunicator.SetChargerQueryPeriod(const Value: integer);
begin
  EnterCriticalSection(csTimeouts);
  fChargerQueryPeriod := Value;
  LeaveCriticalSection(csTimeouts);
end;

procedure TCommunicator.SetChargerWaitAnswer(const Value: integer);
begin
  EnterCriticalSection(csTimeouts);
  fChargerWaitAnswer := Value;
  LeaveCriticalSection(csTimeouts);
end;

procedure TCommunicator.SetFlashLenght(const Value: integer);
begin
  FFlashLenght := Value;

  if (Value mod CFLASH_SIZE_BLOCK) <> 0 then
    raise Exception.Create('Invalid flash size in SetFlashLenght');

 SetLength(FFlashBuffer, Value);
end;

function TCommunicator.GetFlashItem(Index: integer): integer;
begin
  result := FFlashBuffer[Index];
end;

procedure TCommunicator.SetFlashBuffer(Index: integer;
  const Value: integer);
begin
  FFlashBuffer[Index] := Value;
end;

function TCommunicator.GetRxStr: string;
begin
  EnterCriticalSection(csRxStr);
  result := frxstr;
  LeaveCriticalSection(csRxStr);
end;

procedure TCommunicator.OnPacket(Sender: TObject; const str: string);
begin
  RxStr := str;
  SetEvent(hRxComplete);
end;

end.
